﻿unit uHTTPServer;

// Demo version of a simple HTTP server that works with both TCP and UDP protocols
{$IFDEF FPC}
{$mode Delphi}
{$ENDIF}

interface

uses Types, SysUtils, StrUtils,
{$IFDEF FPC}LazUTF8, {$ENDIF}Classes, Generics.Collections, IdGlobal, IdUDPClient, IdTCPClient, IdIOHandlerSocket, IdTCPServer, IdUDPServer, IdContext, IdSocketHandle, IdComponent, IdException,
  IdHTTPServer, IdCustomHTTPServer;

const
  TCPServerPort  = 4280;
  UDPServerPort  = 4281;
  HTTPServerPort = 4282;
  HTML_TEXT      = '<HTML><BODY>Привет!</BODY></HTML>';

type
  THttpUDPServer = class(TIdUDPServer)
  protected
    procedure DoUDPRead(AThread: TIdUDPListenerThread; const AData: TIdBytes; ABinding: TIdSocketHandle); override;
  end;

  THttpServer = class
  private
    IndyTCPServer: TIdTCPServer;
    IndyUDPServer: THttpUDPServer;
    IndyHTTPServer: TIdHTTPServer;
    procedure TCPServerExecute(AContext: TIdContext);
    function GetTCPServerActive: boolean;
    function GetUDPServerActive: boolean;
    function GetHTTPServerActive: boolean;
    procedure HTTPCommandGet(AContext: TIdContext; ARequestInfo: TIdHTTPRequestInfo; AResponseInfo: TIdHTTPResponseInfo);
  public
    constructor Create;
    destructor Destroy; override;
    procedure StartTCPServer;
    procedure StartUDPServer;
    procedure StartHTTPServer;
    procedure StopTCPServer;
    procedure StopUDPServer;
    procedure StopHTTPServer;
    property IsTCPServerActive: boolean read GetTCPServerActive;
    property IsUDPServerActive: boolean read GetUDPServerActive;
    property IsHTTPServerActive: boolean read GetHTTPServerActive;
    function SendTCPCommand(APort: integer): TStringDynArray;
    function SendUDPCommand(APort: integer): TStringDynArray;
  end;

implementation

const
  IO_TIMEOUT = 5000;

procedure THttpUDPServer.DoUDPRead(AThread: TIdUDPListenerThread; const AData: TIdBytes; ABinding: TIdSocketHandle);
var
  iBodyLength: integer;
begin
  iBodyLength := Length(UTF8String(HTML_TEXT));
  ABinding.SendTo(ABinding.PeerIP, ABinding.PeerPort, iBodyLength.ToString, ABinding.IPVersion);
  ABinding.SendTo(ABinding.PeerIP, ABinding.PeerPort, HTML_TEXT, ABinding.IPVersion, IndyTextEncoding_UTF8);
  inherited;
end;

procedure THttpServer.StartTCPServer;
begin
  if not Assigned(IndyTCPServer) then
  begin
    IndyTCPServer := TIdTCPServer.Create;
    with IndyTCPServer do
    begin
      DefaultPort := TCPServerPort;
      OnExecute := TCPServerExecute;
    end;
    IndyTCPServer.Active := True;
  end;
end;

procedure THttpServer.StartUDPServer;
begin
  if not Assigned(IndyUDPServer) then
  begin
    IndyUDPServer := THttpUDPServer.Create;
    with IndyUDPServer do
    begin
      DefaultPort := UDPServerPort;
      ThreadedEvent := True;
      // if False you cant send and receive UDP packets simultaneously
    end;
    IndyUDPServer.Active := True;
  end;
end;

procedure THttpServer.StartHTTPServer;
begin
  if not Assigned(IndyHTTPServer) then
  begin
    IndyHTTPServer := TIdHTTPServer.Create;
    with IndyHTTPServer do
    begin
      DefaultPort := HTTPServerPort;
      OnCommandGet := HTTPCommandGet;
    end;
    IndyHTTPServer.Active := True;
  end;
end;

procedure THttpServer.StopTCPServer;
begin
  FreeAndNil(IndyTCPServer);
end;

procedure THttpServer.StopUDPServer;
begin
  FreeAndNil(IndyUDPServer);
end;

procedure THttpServer.StopHTTPServer;
begin
  FreeAndNil(IndyHTTPServer);
end;

function THttpServer.SendTCPCommand(APort: integer): TStringDynArray;
var
  IdTCPClient: TIdTCPClient;
  Body: string;
  AnsiBody: ansistring;
  s: string;
  ContentLength: integer;
  MemStream: TMemoryStream;
begin
  Result := [];
  IdTCPClient := TIdTCPClient.Create;
  with IdTCPClient do
    try
      try
        Port := APort;
        Host := 'localhost';
        ConnectTimeout := IO_TIMEOUT;
        ReadTimeout := IO_TIMEOUT;
        Connect;

        Body := '';
        ContentLength := 0;
        MemStream := TMemoryStream.Create;
        with IOHandler do
          try
            WriteLn('GET / HTTP/1.1');
            WriteLn('Content-Length: 0');
            WriteLn('Host: localhost');
            WriteLn('');
            repeat
              s := readln(IndyTextEncoding_ASCII);
              if s.StartsWith('Content-Length:') then
                ContentLength := s.Substring(Length('Content-Length:') + 1).ToInteger;
            until s = '';
            if ContentLength > 0 then
            begin
              ReadStream(MemStream, ContentLength, False);
            end;
            // parsing HTTP body
            if ContentLength > 0 then
            begin
              MemStream.Position := 0;
              AnsiBody := '';
              SetLength(AnsiBody, ContentLength);
              MemStream.Read((@AnsiBody[1])^, ContentLength);
              Body := {$IFDEF FPC}AnsiBody{$ELSE} UTF8ToString(AnsiBody){$ENDIF}// no need to convert from UTF-8 in Lazarus
            end;
          finally
            MemStream.Free;
          end;
        with TStringList.Create do
          try
            Text := Body;
            Result := ToStringArray;
          finally
            Free;
          end;
      except
        on e: Exception do
          Result := [e.Message];
      end;
    finally
      Free;
    end;
end;

function THttpServer.SendUDPCommand(APort: integer): TStringDynArray;
var
  IdUDPClient: TIdUDPClient;
  Body: string;
  AnsiBody: ansistring;
  s: string;
  ContentLength: integer;
  MemStream: TMemoryStream;
  IdBytes: TIdBytes;
begin
  Result := [];
  IdUDPClient := TIdUDPClient.Create;
  with IdUDPClient do
    try
      try
        Host := 'localhost';
        Port := APort;

        Body := '';
        ContentLength := 0;
        MemStream := TMemoryStream.Create;
        try
          Send('localhost', UDPServerPort, 'GET / HTTP/1.1');
          s := ReceiveString(IO_TIMEOUT);
          if not s.IsEmpty then
            ContentLength := s.ToInteger;
          if ContentLength > 0 then
          begin
            IdBytes := [];
            SetLength(IdBytes, ContentLength);
            ReceiveBuffer(IdBytes, IO_TIMEOUT);
            MemStream.Position := 0;
            MemStream.Write(pansichar(@IdBytes[0])^, Length(IdBytes));
          end;
          // HTTP body is received

          // parsing HTTP body
          if ContentLength > 0 then
          begin
            MemStream.Position := 0;
            AnsiBody := '';
            SetLength(AnsiBody, ContentLength);
            MemStream.Read((@AnsiBody[1])^, ContentLength);
            Body := {$IFDEF FPC}AnsiBody{$ELSE} UTF8ToString(AnsiBody){$ENDIF}// no need to convert from UTF-8 in Lazarus
          end;
        finally
          MemStream.Free;
        end;

        with TStringList.Create do
          try
            Text := Body;
            Result := ToStringArray;
          finally
            Free;
          end;
      except
        on e: Exception do
          Result := [e.Message];
      end;
    finally
      Free;
    end;
end;

procedure THttpServer.TCPServerExecute(AContext: TIdContext);
var
  iBodyLength: integer;
  s: string;
begin
  with AContext.Connection.Socket do
    try
      repeat
        s := readln(IndyTextEncoding_ASCII);
      until s = '';
      iBodyLength := Length(UTF8String(HTML_TEXT));
      WriteLn('Content-Length: ' + iBodyLength.ToString, IndyTextEncoding_ASCII);
      WriteLn('');
      WriteLn(HTML_TEXT, IndyTextEncoding_UTF8);
    except
      on e: Exception do
        if not(e is EIdConnClosedGracefully) then
          // some question remains why is it triggered (both in Delphi and lazarus), later explore the problem
          raise
    end;
end;

procedure THttpServer.HTTPCommandGet(AContext: TIdContext; ARequestInfo: TIdHTTPRequestInfo; AResponseInfo: TIdHTTPResponseInfo);
begin
  AResponseInfo.Charset := 'utf-8';
  AResponseInfo.ContentText := HTML_TEXT;
end;

function THttpServer.GetTCPServerActive: boolean;
begin
  Result := Assigned(IndyTCPServer) and IndyTCPServer.Active;
end;

function THttpServer.GetUDPServerActive: boolean;
begin
  Result := Assigned(IndyUDPServer) and IndyUDPServer.Active;
end;

function THttpServer.GetHTTPServerActive: boolean;
begin
  Result := Assigned(IndyHTTPServer);
end;

constructor THttpServer.Create;
begin
  inherited;
end;

destructor THttpServer.Destroy;
begin
  StopTCPServer;
  StopUDPServer;
  inherited;
end;

end.
